/* box.c: ascii2xfig box parser and renderer
 *
 * Copyright (C) 2006 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include "defs.h"

static void extract_box(struct grid *grid, int row, int col);
static void extract_box_text(struct grid *grid, struct box *box);
static void markup_box(struct grid *grid, struct box *box);

/*****************************************************************************/
/*
 * find all the boxes in the diagram
 */
void find_boxes(struct grid *grid)
{
	char **lines = grid->lines;
	int row, col;

	/* search on all but the bottom row and RHS column */
	for (row = 0; row < grid->nlines - 2; row++) {
		for (col = 0; col < grid->ncols - 2; col++) {
			if (lines[row][col] == '+')
				extract_box(grid, row, col);
		}
	}
}

/*****************************************************************************/
/*
 * attempt to extract a box from the diagram at a given row and column
 * - a box is a rectangle with a '+' at each corner
 * - the corners are connected by lines without spaces in them
 * - the box goes left and down from the current position
 * - a box interior is at least one line and one row in size
 * - boxes don't overlap
 */
static void extract_box(struct grid *grid, int row, int col)
{
	struct box box, *pbox;
	char **lines = grid->lines;
	int n, m, loop;

	box.top = row;
	box.left = col;

	/* transit down the LHS */
	n = box.top + 1;
continue_LHS_transit:
	//printf("- %d -\n", n);
	for (; n <= grid->nlines - 1; n++) {
		switch (lines[n][box.left]) {
		case '+':
			goto determined_LHS;
		case '|':
		case '-':
			continue;
		default:
			goto not_a_box;
		}
	}

	goto not_a_box;
determined_LHS:
	box.bottom = n;
	n++;

	/* transit acros the bottom */
	m = box.left + 1;
continue_bottom_transit:
	//printf("| %d |\n", m);
	for (; m <= grid->ncols - 1; m++) {
		switch (lines[box.bottom][m]) {
		case '+':
			goto determined_bottom;
		case '|':
		case '-':
			continue;
		default:
			goto continue_LHS_transit;
		}
	}

	goto continue_LHS_transit;
determined_bottom:
	box.right = m;
	m++;

	_debug("Maybe BOX at (%d,%d) to (%d,%d)",
	       box.top, box.left, box.bottom, box.right);

	/* transit up the RHS */
	for (loop = box.bottom - 1; loop > box.top; loop--) {
		switch (lines[loop][box.right]) {
		case '+':
		case '|':
		case '-':
			continue;
		default:
			goto continue_bottom_transit;
		}
	}

	if (loop != box.top) {
		_debug("reject RHS: %d != %d", loop, box.top);
		goto continue_bottom_transit;
	}

	/* transit back along the top */
	for (loop = box.right; loop > box.left; loop--) {
		switch (lines[box.top][loop]) {
		case '+':
		case '|':
		case '-':
			continue;
		default:
			goto continue_bottom_transit;
		}
	}

	if (loop != box.left) {
		_debug("reject top: %d != %d", loop, box.left);
		goto continue_bottom_transit;
	}

	_debug("BOX at (%d,%d) to (%d,%d)",
	       box.top, box.left, box.bottom, box.right);

	pbox = xmalloc(sizeof(*pbox));
	pbox->top = box.top;
	pbox->left = box.left;
	pbox->bottom = box.bottom;
	pbox->right = box.right;
	pbox->text = NULL;
	pbox->link = grid->boxes;
	grid->boxes = pbox;

	extract_box_text(grid, pbox);
	markup_box(grid, pbox);

not_a_box:
	return;
}

/*****************************************************************************/
/*
 * extract the text inside of a box
 */
static void extract_box_text(struct grid *grid, struct box *box)
{
	struct text *text;
	char **lines = grid->lines;
	int row, col, start, len;

	for (row = box->top + 1; row < box->bottom; row++) {
		for (col = box->left + 1; col < box->right; col++) {
			if (lines[row][col] != ' ')
				goto found_line_of_text;
		}

		continue;

	found_line_of_text:
		start = col;
		for (col = box->right - 1; col > start; col--)
			if (lines[row][col] != ' ')
				break;

		len = col - start + 1;
		_debug("TEXT: '%*.*s'", len, len, lines[row] + start);

		text = xmalloc(sizeof(*text) + len + 1);
		text->len = len;
		text->row = row;
		memcpy(text->text, lines[row] + start, len);
		text->text[len] = '\0';
		text->link = box->text;
		box->text = text;
	}
}

/*****************************************************************************/
/*
 * mark the outline of a box
 * - prevent intepretation of box bars as bits of arrow
 */
static void markup_box(struct grid *grid, struct box *box)
{
	char **lines = grid->lines;
	int row, col;

	/* corners */
	grid->markup[box->top   ][box->left ] |= MARK_BOX;
	grid->markup[box->bottom][box->left ] |= MARK_BOX;
	grid->markup[box->bottom][box->right] |= MARK_BOX;
	grid->markup[box->top   ][box->right] |= MARK_BOX;

	/* sides */
	for (row = box->top + 1; row < box->bottom; row++) {
		if (lines[row][box->left] == '|')
			grid->markup[row][box->left] |= MARK_BOX;
		if (lines[row][box->right] == '|')
			grid->markup[row][box->right] |= MARK_BOX;
	}

	/* top and bottom */
	for (col = box->left + 1; col < box->right; col++) {
		if (lines[box->top][col] == '-')
			grid->markup[box->top][col] |= MARK_BOX;
		if (lines[box->bottom][col] == '-')
			grid->markup[box->bottom][col] |= MARK_BOX;
	}

	/* interior */
	for (row = box->top + 1; row < box->bottom; row++)
		for (col = box->left + 1; col < box->right; col++)
			grid->markup[row][col] |= MARK_BOX;
}

/*****************************************************************************/
/*
 * free up the given box
 */
void free_box(struct box *box)
{
	struct text *text;

	while (box->text) {
		text = box->text;
		box->text = text->link;
		free_text(text);
	}
	free(box);
}

/*****************************************************************************/
/*
 * free up the boxes in a grid
 */
void free_boxes(struct grid *grid)
{
	struct box *box;

	/* free up the memory */
	while (grid->boxes) {
		box = grid->boxes;
		grid->boxes = box->link;
		free_box(box);
	}
}

/*****************************************************************************/
/*
 * render a box as XFIG
 */
void render_box(const struct box *box, FILE *out)
{
	const struct text *text;

	fprintf(out,
		"2"		/* object_code: polyline */
		" 2"		/* sub_type: box */
		" 0"		/* line_style: solid */
		" 1"		/* thickness */
		" 0"		/* pen_color: black */
		" 7"		/* fill_color: white */
		" 50"		/* depth */
		" -1"		/* pen_style - not used */
		" 19"		/* area_fill: light grey fill */
		" 0.000"	/* style_val (1/80 inch) */
		" 0"		/* join_style: mitre */
		" 0"		/* cap_style: butt */
		" -1"		/* radius */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 5"		/* npoints */
		"\n");

	_debug("BOX: (%d,%d) -> (%d,%d)",
	       box->top, box->left, box->bottom, box->right);

	fprintf(out, "\t");
	fprintf(out, " %d %d", XSCALEH(box->left),  YSCALEH(box->top));
	fprintf(out, " %d %d", XSCALEH(box->left),  YSCALEH(box->bottom));
	fprintf(out, " %d %d", XSCALEH(box->right), YSCALEH(box->bottom));
	fprintf(out, " %d %d", XSCALEH(box->right), YSCALEH(box->top));
	fprintf(out, " %d %d", XSCALEH(box->left),  YSCALEH(box->top));
	fprintf(out, "\n");

	for (text = box->text; text; text = text->link) {
		fprintf(out,
			"4"		/* object: text */
			" 1"		/* sub_type: centre justified */
			" 0"		/* color: black */
			" 49"		/* depth */
			" -1"		/* pen_style - not used */
			" 0"		/* font: times roman */
			" 10"		/* font_size (pts) */
			" 0.000"	/* angle (radians) */
			" 4"		/* font flags: PS font */
			" %d"		/* height */
			" %d"		/* length */
			" %d"		/* x */
			" %d"		/* y */
			" %s\\001"	/* string */
			"\n",
			2,
			XSCALE(box->right - box->left),
			XSCALEH(box->left) + XHSCALE(box->right - box->left),
			YSCALEH(text->row) + YHSCALE(1),
			text->text
			);
	}
}
