/* text.c: ascii2xfig text blob parser and renderer
 *
 * Copyright (C) 2006 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include "defs.h"

static void extract_text(struct grid *grid, int row, int col);

/*****************************************************************************/
/*
 * find all the pieces of detached text in the diagram
 */
void find_texts(struct grid *grid)
{
	char **lines = grid->lines;
	int row, col;

	for (row = 0; row < grid->nlines; row++) {
		for (col = 0; col < grid->ncols; col++) {
			if (grid->markup[row][col])
				continue;
			if (isalnum(lines[row][col])) {
				if (lines[row][col] == 'V' &&
				    lines[row - 1][col] == '|')
					continue;
				extract_text(grid, row, col);
			}
		}
	}
}

/*****************************************************************************/
/*
 * extract a block of text from the diagram
 * - a block of text starts with a letter or a number
 * - a block of text has all lines starting in the same column
 */
static void extract_text(struct grid *grid, int row, int col)
{
	struct textbox *tb;
	struct text *txt, **pptxt;
	unsigned char *q;
	char **lines = grid->lines, *p;
	int len;

	tb = xmalloc(sizeof(*tb));
	tb->col = col;
	tb->rows = NULL;
	tb->link = grid->texts;
	grid->texts = tb;

	_debug("EXTRACT");

	pptxt = &tb->rows;
	do {
		/* find the end of each row of text */
		for (p = lines[row] + col; *p; p++) {
			switch (*p) {
			case ' ':
				if (p[1] == ' ')
					goto found_end;
				break;

			case '-':
				if (p[1] == '-') {
					p--;
					goto found_end;
				}
				break;

			case '+':
				if (p[1] == '-') {
					p--;
					goto found_end;
				}
				break;

			case '^':
			case '|':
			case '\\':
				p--;
				goto found_end;

			default:
				break;
			}
		}
	found_end:

		/* remove trailing spaces */
		while (p[-1] == ' ')
			p--;

		len = p - (lines[row] + col);
		_debug("- txt: '%*.*s' %c", len, len, lines[row] + col, *p);

		/* save the text */
		txt = xmalloc(sizeof(*txt) + len + 1);
		txt->len = len;
		txt->row = row;
		txt->link = NULL;
		memcpy(txt->text, lines[row] + col, len);
		txt->text[len] = '\0';

		*pptxt = txt;
		pptxt = &txt->link;

		/* make sure we don't see this text again */
		q = grid->markup[row] + col;
		for (; len > 0; len--)
			*q++ |= MARK_TEXT;

	} while (row++, isalnum(lines[row][col]));
}

/*****************************************************************************/
/*
 * free a text box and all its rows of text
 */
static void free_textbox(struct textbox *tb)
{
	struct text *text;

	while (tb->rows) {
		text = tb->rows;
		tb->rows = text->link;
		free_text(text);
	}
	free(tb);
}

/*****************************************************************************/
/*
 * free all the text boxes attached to a grid
 */
void free_texts(struct grid *grid)
{
	struct textbox *tb;

	/* free up the memory */
	while (grid->texts) {
		tb = grid->texts;
		grid->texts = tb->link;
		free_textbox(tb);
	}
}

/*****************************************************************************/
/*
 * render the text in a textbox
 */
void render_text(const struct textbox *tb, FILE *out)
{
	const struct text *text;

	for (text = tb->rows; text; text = text->link) {
		fprintf(out,
			"4"		/* object: text */
			" 0"		/* sub_type: left justified */
			" 0"		/* color: black */
			" 49"		/* depth */
			" -1"		/* pen_style - not used */
			" 0"		/* font: times roman */
			" 10"		/* font_size (pts) */
			" 0.000"	/* angle (radians) */
			" 4"		/* font_flags: PS font */
			" %d"		/* height */
			" %d"		/* length */
			" %d"		/* x */
			" %d"		/* y */
			" %s\\001"	/* string */
			"\n",
			YSCALE(1),
			XSCALE(text->len),
			XSCALE(tb->col),
			YSCALEH(text->row) + YHSCALE(1),
			text->text
			);
	}
}
